
MACRO_DEF(health),
MACRO_DEF(armortype),
MACRO_DEF(armor),
MACRO_DEF(colored_armor),
MACRO_DEF(colored_powerups),
MACRO_DEF(colored_short_powerups),
MACRO_DEF(tp_powerups),
MACRO_DEF(shells),
MACRO_DEF(nails),
MACRO_DEF(rockets),
MACRO_DEF(cells),
MACRO_DEF(weapons),
MACRO_DEF(ammo),
MACRO_DEF(bestweapon),
MACRO_DEF(bestammo),
MACRO_DEF(powerups),
MACRO_DEF(location),
MACRO_DEF(deathloc),
MACRO_DEF(tookatloc),
MACRO_DEF(tookloc),
MACRO_DEF(took),
MACRO_DEF(pointatloc),
MACRO_DEF(pointloc),
MACRO_DEF(point),
MACRO_DEF(need),
MACRO_DEF(droploc),
MACRO_DEF(droptime),
MACRO_DEF(ledpoint),
MACRO_DEF(ledstatus),
MACRO_DEF(lastloc),
MACRO_DEF(lastpowerup),
MACRO_DEF(cam_pos_x),
MACRO_DEF(cam_pos_y),
MACRO_DEF(cam_pos_z),
MACRO_DEF(cam_pos),
MACRO_DEF(cam_angles_pitch),
MACRO_DEF(cam_angles_yaw),
MACRO_DEF(cam_angles_roll),
MACRO_DEF(cam_angles),
MACRO_DEF(demoname),
MACRO_DEF(demolength),
MACRO_DEF(connectiontype),
MACRO_DEF(demoplayback),
MACRO_DEF(demotime),
MACRO_DEF(rand),
MACRO_DEF(matchstatus),
MACRO_DEF(serverip),
MACRO_DEF(conwidth),
MACRO_DEF(conheight),
MACRO_DEF(matchname),
MACRO_DEF(matchtype),
MACRO_DEF(mp3info),
MACRO_DEF(mp3_volume),
MACRO_DEF(lastip),
MACRO_DEF(qt),
MACRO_DEF(latency),
MACRO_DEF(ping),
MACRO_DEF(time),
MACRO_DEF(timestamp),
MACRO_DEF(date),
MACRO_DEF(weaponnum),
MACRO_DEF(weapon),
MACRO_DEF(tf_skin),
MACRO_DEF(gamedir),
MACRO_DEF(triggermatch),
MACRO_DEF(team1),
MACRO_DEF(team2),

// leave trailing ,
