
#ifndef EZQUAKE_R_PERFORMANCE_HEADER
#define EZQUAKE_R_PERFORMANCE_HEADER

void R_PerformanceBeginFrame(void);
void R_PerformanceEndFrame(void);

void GLC_TimeRefresh(void);
void GLM_TimeRefresh(void);

#endif // #ifndef EZQUAKE_R_PERFORMANCE_HEADER