
CMDLINE_DEF(client_nosound, "-nosound"),
CMDLINE_DEF(client_democache, "-democache"),
CMDLINE_DEF(client_norjscripts, "-norjscripts"),
CMDLINE_DEF(client_noscripts, "-noscripts"),
CMDLINE_DEF(client_noindphys, "-noindphys"),
CMDLINE_DEF(client_nomultitexturing, "-nomtex"),
CMDLINE_DEF(client_no_npot_textures, "-nonpot"),
CMDLINE_DEF(client_notriplebuffering, "-no-triple-gl-buffer"),
CMDLINE_DEF(client_noinverselightmaps, "-noinvlmaps"),
CMDLINE_DEF(client_forwardonlyprofile, "-gl-forward-only-profile"),
CMDLINE_DEF(client_detailtrails, "-detailtrails"),
CMDLINE_DEF(client_maximum2textureunits, "-maxtmu2"),
CMDLINE_DEF(client_no24bittextures, "-no24bit"),
CMDLINE_DEF(client_forcetexturereload, "-forcetexturereload"),
CMDLINE_DEF(client_showlibraryerrors, "-showliberrors"),
CMDLINE_DEF(client_ruleset, "-ruleset"),
CMDLINE_DEF(client_particlecount, "-particles"),
CMDLINE_DEF(client_nohardwaretimers, "-nohwtimer"),
CMDLINE_DEF(client_allowmultipleclients, "-allowmultiple"),
CMDLINE_DEF(client_unaccelerated_visuals, "-no-accel-visuals"),
CMDLINE_DEF(client_printopenglextensions, "-gl_ext"),
CMDLINE_DEF(client_gamma, "-gamma"),
CMDLINE_DEF(client_nohardwaregamma, "-nohwgamma"),
CMDLINE_DEF(client_oldgammabehaviour, "-oldgamma"),
CMDLINE_DEF(client_windowedmode, "-window"),
CMDLINE_DEF(client_startwindowed, "-startwindowed"),
CMDLINE_DEF(client_video_frequency, "-freq"),
CMDLINE_DEF(client_video_bpp, "-bpp"),
CMDLINE_DEF(client_video_width, "-width"),
CMDLINE_DEF(client_video_height, "-height"),
CMDLINE_DEF(client_video_displaynumber, "-display"),
CMDLINE_DEF(client_video_conwidth, "-conwidth"),
CMDLINE_DEF(client_video_conheight, "-conheight"),
CMDLINE_DEF(client_video_glsl_renderer, "-glsl-renderer"),
CMDLINE_DEF(client_video_r_debug, "-r-debug"),
CMDLINE_DEF(client_video_r_trace, "-r-trace"),
CMDLINE_DEF(client_nostdinput, "-noconinput"),
CMDLINE_DEF(client_nostdoutput, "-nostdout"),
CMDLINE_DEF(client_nolibpng, "-nolibpng"),
CMDLINE_DEF(client_nolibjpeg, "-nolibjpeg"),
CMDLINE_DEF(client_video_nodesktopres, "-nodesktopres"),
CMDLINE_DEF(client_cd_audio, "-cdaudio"),
CMDLINE_DEF(client_cd_device, "-cddev"),
CMDLINE_DEF(client_noatlas, "-noatlas"),
CMDLINE_DEF(client_verify_glstate, "-r-verify"),
CMDLINE_DEF(client_novao, "-r-novao"),
CMDLINE_DEF(client_nocallback, "-r-nocallback"),
CMDLINE_DEF(client_nomultibind, "-r-nomultibind"),
CMDLINE_DEF(client_no_amd_fix, "-r-no-amd-fix"),

CMDLINE_DEF(filesystem_basedir, "-basedir"),
CMDLINE_DEF(filesystem_nohome, "-nohome"),
CMDLINE_DEF(filesystem_userdir, "-userdir"),
CMDLINE_DEF(filesystem_game, "-game"),
CMDLINE_DEF(filesystem_data, "-data"),

CMDLINE_DEF(host_memory_minimum, "-minmemory"),
CMDLINE_DEF(host_memory_kb, "-heapsize"),
CMDLINE_DEF(host_memory_mb, "-mem"),

CMDLINE_DEF(net_ipaddress, "-ip"),
CMDLINE_DEF(net_clientport, "-clientport"),
CMDLINE_DEF(net_serverport, "-port"),

CMDLINE_DEF(console_buffersize, "-conbufsize"),
CMDLINE_DEF(console_debug, "-condebug"),
CMDLINE_DEF(developer_mode, "-dev"),

CMDLINE_DEF(server_progtype, "-progtype"),
CMDLINE_DEF(server_enablecheats, "-cheats"),
CMDLINE_DEF(server_enablelocalcommand, "-enablelocalcommand"),
CMDLINE_DEF(server_democache_kb, "-democache"),
// leave trailing ,

